/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * eMount
 * http://emount.sourceforge.net/
 * 
 * Copyright (C) Aqueio 2010-2011 <aqueio@yahoo.com>
 * 
 * eMount is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * eMount is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Pline.h"
#include "Filter.h"
#include "../QDebug.h"
#include <assert.h>

#include <glibmm/i18n.h>
#include <iostream>

namespace Pline
{

//==============================================================================

Pline::Pline(   const eDataMode eDataModeItem,
				VolumeConfig *poVolumeConfig)
: goVolumeConfig((*poVolumeConfig))
{
	geDataMode = eDataModeItem;
	gbRun = false;
	gpoReqFilter = 0;
	gbSkipCompletedSignal = false;
			
	initError();
}

//==============================================================================

Pline::~Pline()
{		
	removeFilters();
}

//==============================================================================

void Pline::addFilter(Filter *poFilter)
{	
	// connect filter signals
	poFilter->signal_req_data().connect(
		sigc::bind<1>(sigc::mem_fun(*this, &Pline::onReqData), poFilter));
	
	poFilter->signal_update().connect(
		sigc::bind<1>(sigc::mem_fun(*this, &Pline::onUpdate), poFilter));
		
	gvecFilter.push_back(poFilter);
}

//==============================================================================

void Pline::askData()
{
	assert(gpoReqFilter);
	
	Filter *poReqFilter = gpoReqFilter;
	
	gpoReqFilter = 0;
	poReqFilter->askData();
	
	return;
}

//==============================================================================

std::vector<Filter*> Pline::getFilters()
{
	std::vector<Filter*> vecFilter;
	
	for (std::vector<Filter*>::size_type xi = 0; xi < gvecFilter.size(); ++xi)
		vecFilter.push_back(gvecFilter[xi]);
		
	return vecFilter;
}

//==============================================================================

Pline::tError Pline::getLastError()
{
	if (gtLastError.stErrMsg.empty())
		gtLastError.stErrMsg = _("Unknown error.");
	
	return gtLastError;
}

//==============================================================================

void Pline::initError()
{
	gtLastError.stErrMsg.clear();
	gtLastError.stFilterName.clear();
	gtLastError.bIsFilterError = false;
	gtLastError.bHasCancelledDataReq = false;
	gtLastError.bHasCancelledAction = false;
}

//==============================================================================

bool Pline::isRunning()
{
	Glib::Mutex::Lock oLock(gmtxRun);
	
	return gbRun;
}

//==============================================================================

bool Pline::onFilterPreRun(Filter *poFilter)
{
	return true; // return false would skip this filter from running
}

//==============================================================================

bool Pline::onFilterRun(Filter *poFilter, bool bHasRunWithErrors)
{
	return true; // return false would interrupt pline running if bHasRunWithErrors == false
}
	
//==============================================================================

void Pline::onReqData(Filter *poFilter)
{
	assert(!gpoReqFilter);
	
	gpoReqFilter = poFilter;
	
	(geDataMode == kNoSignalDataMode) ?  askData() : m_signal_req_data();
}

//==============================================================================

void Pline::onUpdate(Filter *poFilter)
{
	if (geDataMode == kNoSignalDataMode)
		return;
	
	m_signal_update();
}

//==============================================================================

void Pline::removeFilters()
{
	std::vector<Filter*> vecUniqueFilter;
	bool bFoundInUnique;

	// find filter instances
	for (std::vector<Filter*>::size_type xi = 0; xi < gvecFilter.size(); ++xi)
	{
		Filter* poFilter = gvecFilter[xi];
		
		// select unique filters
		bFoundInUnique = false;
			
		for (   std::vector<Filter*>::size_type xiUnique = 0; 
				xiUnique < vecUniqueFilter.size();
				++xiUnique)
		{
			if (vecUniqueFilter[xiUnique] == poFilter)
			{
				bFoundInUnique = true;
				break;
			}
		}
			
		if (!bFoundInUnique)
			vecUniqueFilter.push_back(poFilter);
	}

	for (std::vector<Filter*>::size_type xi = 0;  xi < vecUniqueFilter.size(); ++xi)
		delete vecUniqueFilter[xi];
	
	gvecFilter.clear();
}

//==============================================================================

void Pline::reverseFilters()
{
	reverse(gvecFilter.begin(), gvecFilter.end());
}

//==============================================================================

bool Pline::run()
{
	bool	bErr = false;
	bool	bFilterErr = false;
	Filter *poFilter = 0;
	
	gmtxRun.lock();
	gbRun = true;
	gmtxRun.unlock();
	
	initError();

	for (std::vector<Filter*>::size_type xi = 0; xi < gvecFilter.size(); ++xi)
	{
		poFilter = gvecFilter[xi];
		
		// Run filter
		if (!onFilterPreRun(poFilter))
			continue;
		
		// Dump data
#ifdef QDEBUG
		if (poFilter->getSrcData())
		{
			debug_message(DEBUG_PLINE,  "Filter[%s] Src[%s]", 
										poFilter->getFilterName().c_str(),
										poFilter->getSrcData()->getSrc().c_str());
		}
		else
			debug_message(DEBUG_PLINE,  "Filter[%s] Src[null]", 
										poFilter->getFilterName().c_str());
#endif

		if (!poFilter->run())
		{
#ifdef QDEBUG
			if (poFilter->getSrcData())
			{
				debug_message(DEBUG_PLINE,  "Filter[%s] Src[%s] error", 
											poFilter->getFilterName().c_str(),
											poFilter->getSrcData()->getSrc().c_str());
			}
			else
				debug_message(DEBUG_PLINE,  "Filter[%s] Src[null] error", 
											poFilter->getFilterName().c_str());
#endif
			gtLastError.stErrMsg = poFilter->getLastError();
			gtLastError.stFilterName = poFilter->getFilterName();
			gtLastError.bIsFilterError = true;
			gtLastError.bHasCancelledDataReq = poFilter->getHasCancelledDataReq();
			
			onFilterRun(poFilter, true);
		
			bFilterErr = true;
			break;
		}
		else
		{
#ifdef QDEBUG
			if (poFilter->getSrcData())
			{
				debug_message(DEBUG_PLINE,  "Filter[%s] Src[%s] success", 
											poFilter->getFilterName().c_str(),
											poFilter->getSrcData()->getSrc().c_str());
			}
			else
				debug_message(DEBUG_PLINE,  "Filter[%s] Src[null] success", 
											poFilter->getFilterName().c_str());
#endif
		}
		
		m_signal_update();
					
		if (!onFilterRun(poFilter, false))
		{
			bFilterErr = true;
			break;
		}
	}
	
	// set error
	if (bFilterErr)
		bErr = true;

	gmtxRun.lock();
	gbRun = false;
	gmtxRun.unlock();
	
	// signal
	if (!gbSkipCompletedSignal)
		m_signal_completed(!bErr);
										
	return !bErr;
}

//==============================================================================

} // Pline
