/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * eMount
 * http://emount.sourceforge.net/
 * 
 * Copyright (C) Aqueio 2010-2011 <aqueio@yahoo.com>
 * 
 * eMount is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * eMount is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "TrashFilter.h"
#include "../QDebug.h"
#include "../Util.h"
#include "../eMountSys.h"
#include "../VolumeDialog.h"
#include "MountUtil.h"
#include "MountData.h"

#include <glibmm/i18n.h>
#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

namespace Pline
{

//==============================================================================

TrashFilter::TrashFilter()
: Filter()
{
	gpoMountData = 0;
	
	gbEmptyTrash = false;
}

//==============================================================================

TrashFilter::~TrashFilter()
{
}

//==============================================================================

void TrashFilter::askData()
{	
	if (gvecTrash.size() == 0)
		return Filter::askData();
		
	if (eMountSys::getInstance().getRunMode() != eMountSys::kRunGui)
		return Filter::askData();
	else // GUI
	{
		VolumeDialog::tConfig tConfigItem;
		VolumeDialog::initConfig(tConfigItem);
			
		tConfigItem.bAutoRun = false;
		tConfigItem.bWaitData = false;
		tConfigItem.bYesNo = true;
		tConfigItem.bCancelButton = true;
		tConfigItem.bCancelButtonHasFocus = true;
		tConfigItem.stLabel = gpoMountData->getLabel();
		tConfigItem.stSrc = gpoMountData->getSrc();
		tConfigItem.stCommand = getFilterName();
		
		tConfigItem.stMsg = 
			(gvecTrash.size() == 1) ?
				_("Trash found:\n\n") : _("Trashes found:\n\n");
		
		for (	std::vector<Glib::ustring>::size_type xi = 0;
				xi<gvecTrash.size();
				++xi)
		{
			tConfigItem.stMsg.append(Glib::path_get_basename(gvecTrash[xi]));
			tConfigItem.stMsg.append("\n");
		}
		
		tConfigItem.stMsg.append(
			(gvecTrash.size() == 1) ?
				"\nEmpty trash?" : "\nEmpty trashes?");
		
		VolumeDialog oDlg(tConfigItem);
			
		oDlg.setMountData(gpoMountData);
			
		gbEmptyTrash = (oDlg.run() == Gtk::RESPONSE_OK);
					
		oDlg.hide();
			
		while (Gtk::Main::events_pending()) // force redraw
			Gtk::Main::iteration();
	}
		
	Filter::askData();
}

//==============================================================================

bool TrashFilter::emptyTrash()
{
	bool bErr = false;
	
	for (	std::vector<Glib::ustring>::size_type xi = 0;
			xi<gvecTrash.size();
			++xi)
	{
		Glib::ustring stFilename = gvecTrash[xi];
		
		Glib::ustring stCmd = Util::compose("/bin/rm -fr \"%s\"",
											stFilename.c_str());
														
		if (!Util::exec(stCmd))
			bErr = true;
	}
	
	return !bErr;
}
		
//==============================================================================

bool TrashFilter::hasTrash()
{
	return (gvecTrash.size() > 0);
}

//==============================================================================

bool TrashFilter::run()
{
	bool bEmptyTrash = false;

	if (eMountSys::getInstance().getRunMode() != eMountSys::kRunGui)
		return true; // in shell mode filter is disabled
	
	assert(gpoMountData);
	
	// reset error
	gstLastError.clear();
	
	if (!gpoMountData->isEmptyTrash())
		return true;
	
	// read-only fsys skip
	Glib::ustring stFSys = gpoMountData->getFSys();
	
	if (stFSys.empty())
	{
		if (!MountUtil::getFSys(gpoMountData->getSrc(), &stFSys))
		{
			gstLastError = _("Could not empty the trash.");
			return false;
		}
	}
	
	if (stFSys.lowercase() == "iso9660")
		return true;
	
	// look for trashes
	updateTrash();
	
	if (hasTrash())
	{
		// empty trash
		if (gpoMountData->isEmptyTrashConfirm())
		{
			requestData();	

			if (!waitData())
				return false;
			
			bEmptyTrash = gbEmptyTrash;
		}
		else
			bEmptyTrash = true;
	}
	
	if (bEmptyTrash && !emptyTrash())
	{
		gstLastError = _("Could not empty the trash.");
		return false;
	}
	else
		return true;
}

//==============================================================================

void TrashFilter::setSrcData(MountData *poData)
{
	gpoMountData = poData;
	
	Filter::setSrcData(poData);
}

//==============================================================================

void TrashFilter::updateTrash()
{
	assert(gpoMountData);
	
	Glib::Dir dir(gpoMountData->getDir());
	
	std::vector<Glib::ustring> vecEntry(dir.begin(), dir.end());
	
	gvecTrash.clear();
	
	for (	std::vector<Glib::ustring>::size_type xi = 0;
			xi<vecEntry.size();
			++xi)
	{
		Glib::ustring stFilename = vecEntry[xi];
		Glib::ustring stRe = "^\\.[tT]{1}rash-";
		
		Glib::RefPtr<Glib::Regex> re = 
			Glib::Regex::create(stRe);
	
		if (re->match(stFilename))
		{
			gvecTrash.push_back(
				Glib::build_filename(gpoMountData->getDir(), stFilename));
		}
	}
}

//==============================================================================

} // Pline
