/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * eMount
 * http://emount.sourceforge.net/
 * 
 * Copyright (C) Aqueio 2010-2011 <aqueio@yahoo.com>
 * 
 * eMount is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * eMount is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "VolumeMountGui.h"
#include "QDebug.h"
#include "eMountWin.h"
#include "DispatchData.h"
#include "Util.h"

#include <glibmm/i18n.h>
#include <assert.h>

//==============================================================================

VolumeMountGui::VolumeMountGui(VolumeConfig *poVolumeConfig)
: VolumeGui(poVolumeConfig)
{
	bool bMounted = false;
	
	// init
	gpoMountPline = 0;
	gpoProcessPline = 0;
	
	gpoClonePline = 0;
	gpoEnlargeVolumePline = 0;
	gpoInitVolumePline = 0;
	gpoMkDiskImagePline = 0;
	gpoMkVolumePline = 0;
	
	gpoNewPlineVolumeConfig = 0;
	
	gbSigResultCode = false;
	gbSigMount = false;
	gbSigEject = false;
	
	gbMountPlineBusy = false;
	
	gpoMountPline = new Pline::MountPline(  Pline::kWithSignalDataMode,
											&goVolumeConfig);
		
	// update mountpline status
	Glib::Mutex::Lock lock(gmtxMountPline);
	
	gbMountPlineBusy = true;
	updateMountPlineStatus();
	gbMountPlineBusy = false;
	
	// set gui
	updateVolumeGuiState(gpoMountPline->isInUse(&bMounted) && bMounted);
	
	// connect mount pline signals
	connectPlineSignals(gpoMountPline);
}

//==============================================================================

VolumeMountGui::~VolumeMountGui()
{	
	assert(!gpoClonePline);
	assert(!gpoEnlargeVolumePline);
	assert(!gpoInitVolumePline);
	assert(!gpoMkDiskImagePline);
	assert(!gpoMkVolumePline);
	assert(!gpoNewPlineVolumeConfig);
	
	if (gpoMountPline)
		delete gpoMountPline;
}

//==============================================================================

void VolumeMountGui::calcDiskUsage()
{
	Glib::Mutex::Lock lock(gmtxMountPline);

	assert(gpoMountPline);
	
	if (gbMountPlineBusy)
		return;
	
	gpoMountPline->setRunMode(Pline::MountPline::kRunDiskUsage);
	gpoMountPline->run();
}

//==============================================================================

bool VolumeMountGui::cancel()
{
	if (!isRunning())
		return false;
	
	if (!gmtxPline.trylock())
		return false;
	
	if (gpoProcessPline)
	{
		if (!gpoProcessPline->isCanCancel())
		{
			gmtxPline.unlock();
			return false;
		}
		
		gpoProcessPline->cancel();
	}
	
	gmtxPline.unlock();
	
	return true;
}

//==============================================================================

bool VolumeMountGui::clone()
{
	if (!VolumeGui::canStartPline() || !VolumeGui::lockStartPline())
		return false;
	
	gpoClonePline = new Pline::ClonePline(  Pline::kWithSignalDataMode,
											&goVolumeConfig);
		
	gpoProcessPline = gpoClonePline;
	
	connectPlineSignals(gpoClonePline);
			
	// launch thread
	debug_message(DEBUG_ALL, "Starting clone.");
	
	VolumeGui::launchThread();
	
	return false;
}

//==============================================================================

void VolumeMountGui::connectPlineSignals(Pline::Pline *poPline)
{
	// connect MountPline data signal
	poPline->signal_req_data().connect(
			sigc::mem_fun(*this, &VolumeMountGui::onPlineReqData));
	
	poPline->signal_completed().connect(
			sigc::mem_fun(*this, &VolumeMountGui::onPlineRunCompleted));
	
	poPline->signal_update().connect(
			sigc::mem_fun(*this, &VolumeMountGui::onPlineUpdateReq));
}

//==============================================================================

bool VolumeMountGui::createDiskImage(	const unsigned long long ullVolumeSize,
										const bool bUseFullAllocation,
										const Glib::ustring& stCipher,
										const Glib::ustring& stCBC,
										const Glib::ustring& stKeySize)
{
	if (!VolumeGui::canStartPline() || !VolumeGui::lockStartPline())
		return false;
	
	gpoMkDiskImagePline = 
		new Pline::MkDiskImagePline(  	Pline::kWithSignalDataMode,
										&goVolumeConfig,
										ullVolumeSize,
										bUseFullAllocation,
										stCipher,
										stCBC,
										stKeySize);
	
	gpoProcessPline = gpoMkDiskImagePline;
	
	connectPlineSignals(gpoMkDiskImagePline);
	
	// launch thread
	debug_message(DEBUG_ALL, "Starting create disk image.");
	
	VolumeGui::launchThread();
	
	return false;
}

//==============================================================================

bool VolumeMountGui::createVolume(	const Glib::ustring& stCipher,
									const Glib::ustring& stCBC,
									const Glib::ustring& stKeySize)
{
	if (!VolumeGui::canStartPline() || !VolumeGui::lockStartPline())
		return false;
	
	gpoMkVolumePline = 
		new Pline::MkVolumePline(  	Pline::kWithSignalDataMode,
									&goVolumeConfig,
									stCipher,
									stCBC,
									stKeySize);
	
	gpoProcessPline = gpoMkVolumePline;
	
	connectPlineSignals(gpoMkVolumePline);
	
	// launch thread
	debug_message(DEBUG_ALL, "Starting create volume.");
	
	VolumeGui::launchThread();
	
	return false;
}

//==============================================================================

bool VolumeMountGui::eject()
{
	return mountEject(false);
}

//==============================================================================

bool VolumeMountGui::enlargeVolume(const unsigned long long ullEnlargeSize)
{
	if (!VolumeGui::canStartPline() || !VolumeGui::lockStartPline())
		return false;
	
	if (ullEnlargeSize <= 0)
		return false;
		
	gpoEnlargeVolumePline = 
		new Pline::EnlargeVolumePline(  Pline::kWithSignalDataMode,
										&goVolumeConfig,
										ullEnlargeSize);
	
	gpoProcessPline = gpoEnlargeVolumePline;
	
	connectPlineSignals(gpoEnlargeVolumePline);
	
	// launch thread
	debug_message(DEBUG_ALL, "Starting enlarge volume.");
	
	VolumeGui::launchThread();
	
	return false;
}

//==============================================================================

bool VolumeMountGui::initVolume(VolumeConfig* poNewVolumeConfig,
								const Glib::ustring& stCipher,
								const Glib::ustring& stCBC,
								const Glib::ustring& stKeySize)
{
	assert (!gpoNewPlineVolumeConfig);
	
	if (!VolumeGui::canStartPline() || !VolumeGui::lockStartPline())
		return false;
		
	gpoNewPlineVolumeConfig = new VolumeConfig;
	*gpoNewPlineVolumeConfig = *poNewVolumeConfig;
	
	gpoInitVolumePline = 
		new Pline::InitVolumePline( Pline::kWithSignalDataMode,
									gpoNewPlineVolumeConfig,
									&goVolumeConfig,
									stCipher,
									stCBC,
									stKeySize);
	
	gpoProcessPline = gpoInitVolumePline;
	
	// apply changes
	Glib::Mutex::Lock lock(gmtxMountPline);

	assert(gpoMountPline);
	
	gbMountPlineBusy = true;
	goVolumeConfig = *poNewVolumeConfig;
	
	connectPlineSignals(gpoInitVolumePline);
	
	// launch thread
	debug_message(DEBUG_ALL, "Starting intialize volume.");
	
	VolumeGui::launchThread();
	
	return false;
}

//==============================================================================

bool VolumeMountGui::isInUse(bool *pbMounted)
{
	assert(gpoMountPline);
	
	Glib::Mutex::Lock lock(gmtxMountPline);
	
	if (gbMountPlineBusy)
		return true;
		
	return (gpoMountPline->isInUse(pbMounted)) ? true : false;
}

//==============================================================================

bool VolumeMountGui::mount()
{
	return mountEject(true);
}

//==============================================================================

void VolumeMountGui::onJoin()
{	
	VolumeGui::onJoin();
	
	updateVolumeStatus();
	
	if (gbSigMount)
		m_signal_mounted(gbSigResultCode);
	
	gbSigResultCode	= false;
	gbSigEject = false;
	gbSigMount = false;
}

//==============================================================================

void VolumeMountGui::onPlineReqData()
{	
	Glib::Mutex::Lock lock(gmtxMountPline);
	Glib::Mutex::Lock lock2(gmtxPline);
	
	DispatchData *poDispatchData = VolumeGui::createDispatchData();
	
	poDispatchData->setPline((gpoProcessPline) ? (Pline::Pline*)gpoProcessPline : (Pline::Pline*)gpoMountPline);
	poDispatchData->doPlineAskData(true);
	
	eMountWin::queueData(poDispatchData);
	
	goDispatcher.emit();
}

//==============================================================================

void VolumeMountGui::onPlineRunCompleted(const bool bSuccess)
{		
	assert(gpoMountPline);
	
	Glib::Mutex::Lock lock(gmtxMountPline);
	Glib::Mutex::Lock lock2(gmtxPline);
	
	Pline::Pline *poPline = 0;
	
	if (gpoClonePline)
		poPline = gpoClonePline;
	else if (gpoEnlargeVolumePline)
		poPline = gpoEnlargeVolumePline;
	else if (gpoInitVolumePline)
		poPline = gpoInitVolumePline;
	else if (gpoMkDiskImagePline)
		poPline = gpoMkDiskImagePline;
	else if (gpoMkVolumePline)
		poPline = gpoMkVolumePline;
	else
	{
		poPline = gpoMountPline;
		
		gbSigResultCode = bSuccess;
	}
		
	debug_message(	DEBUG_ALL, 
					"Run method completed %s [%s].", 
					(bSuccess) ? "successfully" : "with errors",
					goVolumeConfig.getData().getLabel().c_str());
		
	if (!bSuccess 
		&& !poPline->getLastError().bHasCancelledDataReq
		&& !poPline->getLastError().bHasCancelledAction)
	{
		VolumeDialog::tConfig tVolumeDialogConfigItem;
			
		VolumeDialog::initConfig(tVolumeDialogConfigItem);
			
		tVolumeDialogConfigItem.stLabel = goVolumeConfig.getData().getLabel();
		tVolumeDialogConfigItem.stMsg = poPline->getLastError().stErrMsg;
		tVolumeDialogConfigItem.stCommand = poPline->getLastError().stFilterName;
		
		DispatchData *poDispatchData = VolumeGui::createDispatchData();
			
		poDispatchData->setDialogData(tVolumeDialogConfigItem);
		poDispatchData->doJoin(true);
		poDispatchData->doShowDialog(true);
		
		eMountWin::queueData(poDispatchData);
	}
	else
	{
		DispatchData *poDispatchData = VolumeGui::createDispatchData();
			
		poDispatchData->doJoin(true);
			
		if (poPline != gpoMountPline && bSuccess)
		{			
			VolumeDialog::tConfig tVolumeDialogConfigItem;
			
			VolumeDialog::initConfig(tVolumeDialogConfigItem);
			
			tVolumeDialogConfigItem.stLabel = goVolumeConfig.getData().getLabel();
			tVolumeDialogConfigItem.stMsg = _("Process completed.");
		
			poDispatchData->setDialogData(tVolumeDialogConfigItem);
			poDispatchData->doShowDialog(true);
		}
		
		eMountWin::queueData(poDispatchData);
	}
	
	// mount pline ready
	if (poPline == gpoMountPline)
		gbMountPlineBusy = false;
		
	// delete non-mount pline
	if (gpoClonePline)
	{
		delete gpoClonePline;
		gpoClonePline = 0;
		gpoProcessPline = 0;
	}
	else if (gpoEnlargeVolumePline)
	{
		delete gpoEnlargeVolumePline;
		gpoEnlargeVolumePline = 0;
		gpoProcessPline = 0;
		
		gbMountPlineBusy = true;
		updateMountPlineStatus();
		gbMountPlineBusy = false;
	}
	else if (gpoInitVolumePline)
	{
		delete gpoInitVolumePline;
		gpoInitVolumePline = 0;
		gpoProcessPline = 0;
		
		assert(gpoNewPlineVolumeConfig);
		
		delete gpoNewPlineVolumeConfig;
		gpoNewPlineVolumeConfig = 0;
		
		gbMountPlineBusy = true;
		updateMountPlineStatus();
		gbMountPlineBusy = false;
	}
	else if (gpoMkDiskImagePline)
	{
		delete gpoMkDiskImagePline;
		gpoMkDiskImagePline = 0;
		gpoProcessPline = 0;
	}
	else if (gpoMkVolumePline)
	{
		delete gpoMkVolumePline;
		gpoMkVolumePline = 0;
		gpoProcessPline = 0;
	}
			
	goDispatcher.emit();
}

//==============================================================================

void VolumeMountGui::onPlineUpdateReq()
{
	if (!gmtxPline.trylock())
		return;
			
	if (gpoProcessPline)
	{
		DispatchData *poDispatchData = VolumeGui::createDispatchData();
		
		poDispatchData->doUpdateVolume(true);
			
		eMountWin::queueData(poDispatchData);
	}
	
	gmtxPline.unlock();
	
	goDispatcher.emit();
}

//==============================================================================

void VolumeMountGui::onUpdateVolume()
{
	updateVolumeStatus();
}

//==============================================================================

bool VolumeMountGui::mountEject(const bool bReqMount)
{	
	Glib::Mutex::Lock lock(gmtxMountPline);
	
	gbMountPlineBusy = true;
	
	bool bGuiMount = false;
	bool bDoEject = false;
	bool bDoMount = false;
	
	// warnings
	bool bVolumeAlreadyMounted = false;
	bool bVolumeAlreadyEjected = false;
	bool bVolumeInUseNotMounted = false;
	bool bVolumeMismatchGui = false;
	
	// check status and set Eject or Mount or ask to user
	assert(gpoVolumeGuiState);
	assert(gpoMountPline);
	
	// check if thread is running
	if (!VolumeGui::canStartPline())
		return false;

	bool bPlineMounted = false;
	bool bPlineInUse = gpoMountPline->isInUse(&bPlineMounted);
		
	// set bGuiMount
	bGuiMount = (gpoVolumeGuiState->isMountMode()) ? true : false;
	
	if (!bPlineMounted && bGuiMount)
		bDoMount = true;
	else if (bPlineMounted && !bGuiMount)
		bDoEject = true;
	else if (bPlineMounted && bGuiMount)
		bVolumeMismatchGui = true;
	else if (!bPlineMounted && !bGuiMount)
		bVolumeMismatchGui = true;

	if (bDoMount && !bReqMount)
		bVolumeAlreadyEjected = true;
	else if (bDoEject && bReqMount)
		bVolumeAlreadyMounted = true;
		
	if (bDoMount && bPlineInUse && !bPlineMounted)
		bVolumeInUseNotMounted = true;
	
	// show warnings
	if (bVolumeAlreadyMounted)
	{
		VolumeDialog::tConfig tConfigItem;
		VolumeDialog::initConfig(tConfigItem);
			
		tConfigItem.bAutoRun = true;
		tConfigItem.stLabel = goVolumeConfig.getData().getLabel();
		tConfigItem.stMsg = _("Volume already mounted.");
		
		VolumeDialog oDlg(tConfigItem);
		
		updateMountPlineStatus();
		updateVolumeGuiState(true);
		
		gbMountPlineBusy = false;
		
		return false;
	}
	else if (bVolumeAlreadyEjected)
	{
		VolumeDialog::tConfig tConfigItem;
		VolumeDialog::initConfig(tConfigItem);
			
		tConfigItem.bAutoRun = true;
		tConfigItem.stLabel = goVolumeConfig.getData().getLabel();
		tConfigItem.stMsg = _("Volume already ejected.");
		
		VolumeDialog oDlg(tConfigItem);
		
		updateVolumeGuiState(false);
		
		gbMountPlineBusy = false;
		
		return false;
	}
	else if (bVolumeInUseNotMounted)
	{
		VolumeDialog::tConfig tConfigItem;
		VolumeDialog::initConfig(tConfigItem);
			
		tConfigItem.bAutoRun = false;
		tConfigItem.bCancelButton = true;
		tConfigItem.bCancelButtonHasFocus = false;
		tConfigItem.stLabel = goVolumeConfig.getData().getLabel();
		tConfigItem.stMsg = _("Volume is in use and not properly mounted, eject volume?");
		
		VolumeDialog oDlg(tConfigItem);
		
		if (oDlg.run() == Gtk::RESPONSE_OK)
		{
			bDoMount = false;
			bDoEject = true;
		}
		else
		{
			gbMountPlineBusy = false;
			
			updateVolumeGuiState(false);
			
			return false;
		}
	}
	else if (bVolumeMismatchGui)
	{
		// Volume status updated
		updateMountPlineStatus();
		updateVolumeGuiState(bPlineMounted);
		
		gbMountPlineBusy = false;
		
		return false;
	}
	
	if (!VolumeGui::lockStartPline())
	{
		gbMountPlineBusy = false;
		
		return false;
	}
	
	// run
	if (bDoMount)
	{
		gpoMountPline->setRunMode(Pline::MountPline::kRunMount);
		
		gbSigEject = false;
		gbSigMount = true;
	}
	else if (bDoEject)
	{
		gpoMountPline->setRunMode(Pline::MountPline::kRunEject);
		
		gbSigEject = true;
		gbSigMount = false;
	}
	else
	{
		gbMountPlineBusy = false;
		
		return false;
	}

	// launch thread
	debug_message(DEBUG_ALL, "Starting mounting[%d]/ejecting[%d].", bDoMount, bDoEject);
	
	VolumeGui::launchThread();
		
	return true;
}

//==============================================================================

void VolumeMountGui::threadFun()
{
	assert(gpoMountPline);
	
	//
	// non-mount pline
	//
	if (gpoProcessPline)
	{
		debug_message(DEBUG_ALL, "ProcessPline Thread started.");
		
		gpoProcessPline->run();
				
		return;
	}
	
	//
	// mount pline
	//
	debug_message(DEBUG_ALL, "MountPline Thread started.");
	
	gpoMountPline->run();
}

//==============================================================================

void VolumeMountGui::updateMountPlineStatus()
{
	assert(gbMountPlineBusy);
	
	gpoMountPline->setRunMode(Pline::MountPline::kRunDiskUsage);
	gpoMountPline->run();
		
	gpoMountPline->setRunMode(Pline::MountPline::kRunDumpStatus);
	gpoMountPline->run();
}

//==============================================================================

void VolumeMountGui::updateVolumeGuiState(const bool bPlineMounted)
{
	assert(gpoMountPline);

	// update non-mount pline
	if (gpoProcessPline)
	{
		gpoVolumeGuiState->setProgressMode(
				gpoProcessPline->getPctDone(), 
				gpoProcessPline->isCanCancel(), 
				gpoProcessPline->getInfoMessage()
			);
			
		return;
	}
	
	// update mount pline
	gvecPlineStatus.clear();
	gstFSys.clear();
	gstCipher.clear();
	gstDiskSize.clear();
	gstDiskUsedSpace.clear();
	gstDiskAvailableSpace.clear();
	
	if (!bPlineMounted)
	{
		gpoVolumeGuiState->setMountMode();
		
		gvecPlineStatus = gpoMountPline->getDumpStatus();
		
		return;
	}
		
	Pline::MountPline::tDiskUsage tDiskUsageItem =
		gpoMountPline->getDiskUsage();
			
	if (gpoMountPline->getFSys() == "iso9660")
	{
		if (tDiskUsageItem.bValid)
		{
			gpoVolumeGuiState->setEjectMode(
				0, 
				true, 
				Util::compose(	_("ISO Image %s used."), 
								tDiskUsageItem.stSpaceUsed.c_str())
			);
			
			gstDiskSize = tDiskUsageItem.stDiskSize;
			gstDiskUsedSpace = tDiskUsageItem.stSpaceUsed;
			gstDiskAvailableSpace = tDiskUsageItem.stSpaceAvailable;
		}
		else
		{
			gpoVolumeGuiState->setEjectMode(
				0, 
				true, 
				_("ISO Image.")
			);
		}
	}
	else
	{
		if (tDiskUsageItem.bValid)
		{
			gpoVolumeGuiState->setEjectMode(
				tDiskUsageItem.iUsedSpacePct, 
				false, 
				Util::compose( 	_("%s / %s (%d%%) used, %s free. %s %s."), 
								tDiskUsageItem.stSpaceUsed.c_str(),
								tDiskUsageItem.stDiskSize.c_str(),
								tDiskUsageItem.iUsedSpacePct,
								tDiskUsageItem.stSpaceAvailable.c_str(),
								(gpoMountPline->isEncrypted()) ? _("Encrypted") : _("Un-encrypted"),
								gpoMountPline->getFSys().c_str())
			);
			
			gstDiskSize = tDiskUsageItem.stDiskSize;
			gstDiskUsedSpace = tDiskUsageItem.stSpaceUsed;
			gstDiskAvailableSpace = tDiskUsageItem.stSpaceAvailable;
		}
		else
		{
			gpoVolumeGuiState->setEjectMode(
				tDiskUsageItem.iUsedSpacePct, 
				false, 
				Util::compose( 	_("%s %s."), 
								(gpoMountPline->isEncrypted()) ? _("Encrypted") : _("Un-encrypted"),
								(gpoMountPline->getFSys().empty()) ? _("Unknown") : gpoMountPline->getFSys().c_str())

			);
		}
	}
	
	gvecPlineStatus = gpoMountPline->getDumpStatus();
	gstFSys = gpoMountPline->getFSys();
	gstCipher = gpoMountPline->getCipher();
}

//==============================================================================

void VolumeMountGui::updatePline()
{
	assert(gpoMountPline);
	
	if (!gmtxMountPline.trylock())
		return;
	
	if (!gmtxPline.trylock())
	{
		gmtxMountPline.unlock();
		return;
	}
	
	gpoMountPline->updatePline();
	
	gmtxMountPline.unlock();
	gmtxPline.unlock();
}

//==============================================================================

void VolumeMountGui::updateVolumeStatus()
{
	bool bMounted = false;

	assert(gpoMountPline);
	
	if (!gmtxMountPline.trylock())
		return;
	
	if (!gmtxPline.trylock())
	{
		gmtxMountPline.unlock();
		return;
	}
		
	(gpoProcessPline) ?
		updateVolumeGuiState(false)
		: updateVolumeGuiState(gpoMountPline->isInUse(&bMounted) && bMounted);
	
	gmtxMountPline.unlock();
	gmtxPline.unlock();
	
	eMountWin::updateVolumeGuiState(gpoVolumeGuiState);
}

//==============================================================================
